"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const cli_1 = require("@yarnpkg/cli");
const core_1 = require("@yarnpkg/core");
const core_2 = require("@yarnpkg/core");
const core_3 = require("@yarnpkg/core");
const clipanion_1 = require("clipanion");
const get_1 = (0, tslib_1.__importDefault)(require("lodash/get"));
const set_1 = (0, tslib_1.__importDefault)(require("lodash/set"));
const unset_1 = (0, tslib_1.__importDefault)(require("lodash/unset"));
const Constraints_1 = require("../Constraints");
// eslint-disable-next-line arca/no-default-export
class ConstraintsCheckCommand extends cli_1.BaseCommand {
    constructor() {
        super(...arguments);
        this.fix = clipanion_1.Option.Boolean(`--fix`, false, {
            description: `Attempt to automatically fix unambiguous issues, following a multi-pass process`,
        });
    }
    async execute() {
        const configuration = await core_1.Configuration.find(this.context.cwd, this.context.plugins);
        const { project } = await core_1.Project.find(configuration, this.context.cwd);
        const constraints = await Constraints_1.Constraints.find(project);
        const report = await core_2.StreamReport.start({
            configuration,
            stdout: this.context.stdout,
        }, async (report) => {
            let allSaves = new Set();
            let errors = [];
            for (let t = 0, T = this.fix ? 10 : 1; t < T; ++t) {
                errors = [];
                const result = await constraints.process();
                const modifiedDependencies = new Set();
                await processDependencyConstraints(modifiedDependencies, errors, result.enforcedDependencies, {
                    fix: this.fix,
                    configuration,
                });
                // Dependency constraints work on the manifess, field constraints work on the raw JSON objects
                for (const { manifest } of modifiedDependencies) {
                    const newManifest = {};
                    manifest.exportTo(newManifest);
                    manifest.raw = newManifest;
                }
                const modifiedFields = new Set();
                await processFieldConstraints(modifiedFields, errors, result.enforcedFields, {
                    fix: this.fix,
                    configuration,
                });
                // Persist changes made by the constraints back to the Manifest instance
                for (const { manifest } of modifiedFields)
                    manifest.load(manifest.raw);
                allSaves = new Set([
                    ...allSaves,
                    ...modifiedDependencies,
                    ...modifiedFields,
                ]);
                // If we didn't apply any change then we can exit the loop
                if (modifiedDependencies.size === 0 && modifiedFields.size === 0) {
                    break;
                }
            }
            // save all modified manifests
            await Promise.all([...allSaves].map(async (workspace) => {
                await workspace.persistManifest();
            }));
            // report all outstanding errors
            for (const [messageName, message] of errors) {
                report.reportError(messageName, message);
            }
        });
        if (report.hasErrors())
            return report.exitCode();
        return 0;
    }
}
exports.default = ConstraintsCheckCommand;
ConstraintsCheckCommand.paths = [
    [`constraints`],
];
ConstraintsCheckCommand.usage = clipanion_1.Command.Usage({
    category: `Constraints-related commands`,
    description: `check that the project constraints are met`,
    details: `
      This command will run constraints on your project and emit errors for each one that is found but isn't met. If any error is emitted the process will exit with a non-zero exit code.

      If the \`--fix\` flag is used, Yarn will attempt to automatically fix the issues the best it can, following a multi-pass process (with a maximum of 10 iterations). Some ambiguous patterns cannot be autofixed, in which case you'll have to manually specify the right resolution.

      For more information as to how to write constraints, please consult our dedicated page on our website: https://yarnpkg.com/features/constraints.
    `,
    examples: [[
            `Check that all constraints are satisfied`,
            `yarn constraints`,
        ], [
            `Autofix all unmet constraints`,
            `yarn constraints --fix`,
        ]],
});
async function processDependencyConstraints(toSave, errors, enforcedDependencies, { configuration, fix }) {
    const allIdents = new Map();
    const byWorkspaces = new Map();
    for (const { workspace, dependencyIdent, dependencyRange, dependencyType } of enforcedDependencies) {
        let byWorkspacesStore = byWorkspaces.get(workspace);
        if (typeof byWorkspacesStore === `undefined`)
            byWorkspaces.set(workspace, byWorkspacesStore = new Map());
        let byIdentStore = byWorkspacesStore.get(dependencyIdent.identHash);
        if (typeof byIdentStore === `undefined`)
            byWorkspacesStore.set(dependencyIdent.identHash, byIdentStore = new Map());
        let byDependencyTypeStore = byIdentStore.get(dependencyType);
        if (typeof byDependencyTypeStore === `undefined`)
            byIdentStore.set(dependencyType, byDependencyTypeStore = new Set());
        allIdents.set(dependencyIdent.identHash, dependencyIdent);
        byDependencyTypeStore.add(dependencyRange);
    }
    for (const [workspace, byWorkspacesStore] of byWorkspaces) {
        for (const [identHash, byIdentStore] of byWorkspacesStore) {
            const dependencyIdent = allIdents.get(identHash);
            if (typeof dependencyIdent === `undefined`)
                throw new Error(`Assertion failed: The ident should have been registered`);
            for (const [dependencyType, byDependencyTypeStore] of byIdentStore) {
                // If any of the expected ranges are `null` then the dependency should be removed
                const expectedRanges = byDependencyTypeStore.has(null) ? [null] : [...byDependencyTypeStore];
                if (expectedRanges.length > 2) {
                    errors.push([core_2.MessageName.CONSTRAINTS_AMBIGUITY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} via conflicting ranges ${expectedRanges.slice(0, -1).map(expectedRange => core_3.structUtils.prettyRange(configuration, String(expectedRange))).join(`, `)}, and ${core_3.structUtils.prettyRange(configuration, String(expectedRanges[expectedRanges.length - 1]))} (in ${dependencyType})`]);
                }
                else if (expectedRanges.length > 1) {
                    errors.push([core_2.MessageName.CONSTRAINTS_AMBIGUITY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} via conflicting ranges ${core_3.structUtils.prettyRange(configuration, String(expectedRanges[0]))} and ${core_3.structUtils.prettyRange(configuration, String(expectedRanges[1]))} (in ${dependencyType})`]);
                }
                else {
                    const dependencyDescriptor = workspace.manifest[dependencyType].get(dependencyIdent.identHash);
                    const [expectedRange] = expectedRanges;
                    if (expectedRange !== null) {
                        if (!dependencyDescriptor) {
                            if (fix) {
                                workspace.manifest[dependencyType].set(dependencyIdent.identHash, core_3.structUtils.makeDescriptor(dependencyIdent, expectedRange));
                                toSave.add(workspace);
                            }
                            else {
                                errors.push([core_2.MessageName.CONSTRAINTS_MISSING_DEPENDENCY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} (via ${core_3.structUtils.prettyRange(configuration, expectedRange)}), but doesn't (in ${dependencyType})`]);
                            }
                        }
                        else if (dependencyDescriptor.range !== expectedRange) {
                            if (fix) {
                                workspace.manifest[dependencyType].set(dependencyIdent.identHash, core_3.structUtils.makeDescriptor(dependencyIdent, expectedRange));
                                toSave.add(workspace);
                            }
                            else {
                                errors.push([core_2.MessageName.CONSTRAINTS_INCOMPATIBLE_DEPENDENCY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must depend on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} via ${core_3.structUtils.prettyRange(configuration, expectedRange)}, but uses ${core_3.structUtils.prettyRange(configuration, dependencyDescriptor.range)} instead (in ${dependencyType})`]);
                            }
                        }
                    }
                    else {
                        if (dependencyDescriptor) {
                            if (fix) {
                                workspace.manifest[dependencyType].delete(dependencyIdent.identHash);
                                toSave.add(workspace);
                            }
                            else {
                                errors.push([core_2.MessageName.CONSTRAINTS_EXTRANEOUS_DEPENDENCY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} has an extraneous dependency on ${core_3.structUtils.prettyIdent(configuration, dependencyIdent)} (in ${dependencyType})`]);
                            }
                        }
                    }
                }
            }
        }
    }
}
async function processFieldConstraints(toSave, errors, enforcedFields, { configuration, fix }) {
    const byWorkspaces = new Map();
    for (const { workspace, fieldPath, fieldValue } of enforcedFields) {
        const byWorkspacesStore = core_1.miscUtils.getMapWithDefault(byWorkspaces, workspace);
        const byPathStore = core_1.miscUtils.getSetWithDefault(byWorkspacesStore, fieldPath);
        byPathStore.add(fieldValue);
    }
    for (const [workspace, byWorkspacesStore] of byWorkspaces) {
        for (const [fieldPath, byPathStore] of byWorkspacesStore) {
            const expectedValues = [...byPathStore];
            if (expectedValues.length > 2) {
                errors.push([core_2.MessageName.CONSTRAINTS_AMBIGUITY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${core_3.formatUtils.pretty(configuration, fieldPath, `cyan`)} set to conflicting values ${expectedValues.slice(0, -1).map(expectedValue => core_3.formatUtils.pretty(configuration, String(expectedValue), `magenta`)).join(`, `)}, or ${core_3.formatUtils.pretty(configuration, String(expectedValues[expectedValues.length - 1]), `magenta`)}`]);
            }
            else if (expectedValues.length > 1) {
                errors.push([core_2.MessageName.CONSTRAINTS_AMBIGUITY, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${core_3.formatUtils.pretty(configuration, fieldPath, `cyan`)} set to conflicting values ${core_3.formatUtils.pretty(configuration, String(expectedValues[0]), `magenta`)} or ${core_3.formatUtils.pretty(configuration, String(expectedValues[1]), `magenta`)}`]);
            }
            else {
                const actualValue = (0, get_1.default)(workspace.manifest.raw, fieldPath);
                const [expectedValue] = expectedValues;
                if (expectedValue !== null) {
                    if (actualValue === undefined) {
                        if (fix) {
                            await setWorkspaceField(workspace, fieldPath, expectedValue);
                            toSave.add(workspace);
                        }
                        else {
                            errors.push([core_2.MessageName.CONSTRAINTS_MISSING_FIELD, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${core_3.formatUtils.pretty(configuration, fieldPath, `cyan`)} set to ${core_3.formatUtils.pretty(configuration, String(expectedValue), `magenta`)}, but doesn't`]);
                        }
                    }
                    else if (JSON.stringify(actualValue) !== expectedValue) {
                        if (fix) {
                            await setWorkspaceField(workspace, fieldPath, expectedValue);
                            toSave.add(workspace);
                        }
                        else {
                            errors.push([core_2.MessageName.CONSTRAINTS_INCOMPATIBLE_FIELD, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} must have a field ${core_3.formatUtils.pretty(configuration, fieldPath, `cyan`)} set to ${core_3.formatUtils.pretty(configuration, String(expectedValue), `magenta`)}, but is set to ${core_3.formatUtils.pretty(configuration, JSON.stringify(actualValue), `magenta`)} instead`]);
                        }
                    }
                }
                else {
                    if (actualValue !== undefined && actualValue !== null) {
                        if (fix) {
                            await setWorkspaceField(workspace, fieldPath, null);
                            toSave.add(workspace);
                        }
                        else {
                            errors.push([core_2.MessageName.CONSTRAINTS_EXTRANEOUS_FIELD, `${core_3.structUtils.prettyWorkspace(configuration, workspace)} has an extraneous field ${core_3.formatUtils.pretty(configuration, fieldPath, `cyan`)} set to ${core_3.formatUtils.pretty(configuration, JSON.stringify(actualValue), `magenta`)}`]);
                        }
                    }
                }
            }
        }
    }
}
async function setWorkspaceField(workspace, fieldPath, value) {
    if (value === null) {
        (0, unset_1.default)(workspace.manifest.raw, fieldPath);
    }
    else {
        (0, set_1.default)(workspace.manifest.raw, fieldPath, JSON.parse(value));
    }
}
